# -*- coding: UTF-8 -*-
# vim: fileencoding=UTF-8 filetype=python ff=unix et ts=4 sw=4 sts=4 tw=120
# author: Christer Sjöholm -- hcs AT furuvik DOT net

import os
import pathlib
import sys
from os.path import realpath, expanduser
import configparser
import logging

import oauth2client.client
import xdg

from goobook.storage import Storage

log = logging.getLogger(__name__)  # pylint: disable=invalid-name
LEGACY_CONFIG_FILE = pathlib.Path('~/.goobookrc').expanduser()
LEGACY_AUTH_FILE = pathlib.Path('~/.goobook_auth.json').expanduser()
LEGACY_CACHE_FILE = pathlib.Path('~/.goobook_cache').expanduser()

TEMPLATE = '''\
# Use this template to create your ~/.goobookrc

# "#" or ";" at the start of a line makes it a comment.

[DEFAULT]
# The following are optional, defaults are shown when not other specified.

# This file is written by the oauth library, and should be kept secure,
# it's like a password to your google contacts.
# default is to place it in the XDG_DATA_HOME
;oauth_db_filename: ~/.goobook_auth.json

;cache_filename: ~/.goobook_cache   # default is in the XDG_CACHE_HOME
;cache_expiry_hours: 24
;filter_groupless_contacts: yes

# New contacts will be added to this group in addition to "My Contacts"
# Note that the group has to already exist on google or an error will occur.
# One use for this is to add new contacts to an "Unsorted" group, which can
# be sorted easier than all of "My Contacts".
;default_group:
'''


def read_config(config_file=None):
    """Reads the ~/.goobookrc and any authentication data.

    returns the configuration as a dictionary.

    """
    config = Storage({  # Default values
        'cache_filename': None,
        'oauth_db_filename': None,
        'cache_expiry_hours': '24',
        'filter_groupless_contacts': True,
        'default_group': ''})

    # Search for config file to use
    if config_file:  # config file explicitly given on the commandline
        config_file = os.path.expanduser(config_file)
    else:  # search for goobookrc in XDG dirs and homedir
        config_files = [dir_ / "goobookrc" for dir_ in [xdg.XDG_CONFIG_HOME] +
                        xdg.XDG_CONFIG_DIRS] + [LEGACY_CONFIG_FILE]
        log.debug("config file search path: %s", config_files)
        for config_file_ in config_files:
            if config_file_.exists():
                config_file = str(config_file_)
                log.debug("found config file: %s", config_file)
                break
        else:
            log.debug("no config file found")
            config_file = None
        # else:  # .goobookrc in home directory
        #     config_file = os.path.expanduser(CONFIG_FILE)

    if config_file:
        parser = _get_config(config_file)
    else:
        parser = None

    if parser:
        config.get_dict().update(dict(parser.items('DEFAULT', raw=True)))
        # Handle not string fields
        if parser.has_option('DEFAULT', 'filter_groupless_contacts'):
            config.filter_groupless_contacts = parser.getboolean('DEFAULT', 'filter_groupless_contacts')

    if "client_secret_filename" in config:
        print("WARNING: setting client_secret_filename in {} is deprecated".format(config_file), file=sys.stderr)

    # Search for cache file to use
    if config.cache_filename:  # If explicitly specified in config file
        config.cache_filename = realpath(expanduser(config.cache_filename))
    else:  # search for goobook_cache in XDG dirs and homedir
        cache_files = [xdg.XDG_CACHE_HOME / "goobook_cache", LEGACY_CACHE_FILE]
        log.debug("cache file search path: %s", cache_files)
        for cache_file in cache_files:
            cache_file = cache_file.resolve()
            if cache_file.exists():
                log.debug("found cache file: %s", cache_file)
                break
        else:  # If there is none, create in XDG_CACHE_HOME
            cache_file = xdg.XDG_CACHE_HOME / "goobook_cache"
            log.debug("no cache file found, will use %s", cache_file)
        config.cache_filename = str(cache_file)

    # Search for auth file to use
    if config.oauth_db_filename:  # If explicitly specified in config file
        config.oauth_db_filename = realpath(expanduser(config.oauth_db_filename))
        auth_file = pathlib.Path(config.oauth_db_filename)
    else:  # search for goobook_auth.json in XDG dirs and homedir
        auth_files = [dir_ / "goobook_auth.json" for dir_ in [xdg.XDG_DATA_HOME] +
                      xdg.XDG_DATA_DIRS] + [LEGACY_AUTH_FILE]
        log.debug("auth file search path: %s", auth_files)
        for auth_file in auth_files:
            auth_file = auth_file.resolve()
            if auth_file.exists():
                log.debug("found auth file: %s", auth_file)
                break
        else:  # If there is none, create in XDG_DATA_HOME
            auth_file = xdg.XDG_DATA_HOME / "goobook_auth.json"
            log.debug("no auth file found, will use %s", auth_file)
        config.oauth_db_filename = str(auth_file)

    config.store = oauth2client.file.Storage(config.oauth_db_filename)

    config.creds = config.store.get() if auth_file.exists() else None

    log.debug(config)
    return config


def _get_config(config_file):
    """find, read and parse configuraton."""
    parser = configparser.SafeConfigParser()
    if os.path.lexists(config_file):
        try:
            log.info('Reading config: %s', config_file)
            inp = open(config_file)
            parser.read_file(inp)
            return parser
        except (IOError, configparser.ParsingError) as err:
            raise ConfigError("Failed to read configuration %s\n%s" % (config_file, err))
    return None


class ConfigError(Exception):
    pass



# vim: fileencoding=UTF-8 filetype=python ff=unix expandtab sw=4 sts=4 tw=120
# maintainer: Christer Sjöholm -- goobook AT furuvik DOT net
# authors: Marcus Nitzschke -- marcus.nitzschke AT gmx DOT com
#
# Copyright (C) 2009  Carlos José Barroso
# Copyright (C) 2010  Christer Sjöholm
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The idea is make an interface to google contacts that mimics the behaviour of abook for mutt."""
import collections
import datetime
import email.parser
import email.header
import logging
import os
import pickle
import re
import sys
import time

import httplib2
from googleapiclient.discovery import build

from goobook.storage import Storage, storageify, unstorageify

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

CACHE_FORMAT_VERSION = '5.1'

TypedValue = collections.namedtuple('TypedValue', ['value', 'type'])


class GooBook():
    """Application logic.

    This class can't be used as a library as it looks now, it uses sys.stdin
    print and sys.exit().
    """

    def __init__(self, config):
        self.__config = config
        self.cache = Cache(config)
        self.cache.load()

    def query(self, query, simple=False):
        """Do the query, and print it out in specified format.

        simple=False is the mutt format
        simple format is:
          "Name <email>"
        simple format for group matches is a , separated list of Name <email>
        the group name is not printed using simple format.
        """
        # query contacts
        matching_contacts = sorted(self.__query_contacts(query), key=lambda c: c.display_name)
        # query groups
        matching_groups = sorted(self.__query_groups(query), key=lambda g: g[0])
        if not simple:
            # mutt's query_command expects the first line to be a message,
            # which it discards.
            print('')
        for contact in matching_contacts:
            if contact.emails:
                emailaddrs = sorted(contact.emails)
                groups = set(contact.groups)
                groups = groups.difference(('My Contacts',))
                groups_str = ', '.join(('"' + g + '"' for g in groups))
                for (emailaddr, kind) in emailaddrs:
                    extra_str = kind
                    if groups_str:
                        extra_str = extra_str + ' groups: ' + groups_str
                    if simple:
                        print(f"{contact.display_name} <{emailaddr}>")
                    else:
                        print('\t'.join((emailaddr, contact.display_name, extra_str)))

        for title, contacts in matching_groups:
            emails = ['%s <%s>' % (c.display_name, c.emails[0][0]) for c in contacts if c.emails]
            emails = ', '.join(emails)
            if not emails:
                continue
            if simple:
                print(emails)
            else:
                print('%s\t%s (group)' % (emails, title))

    def query_details(self, query):
        """Method for querying the contacts and printing a detailed view."""
        out = sys.stdout

        # query contacts
        matching_contacts = list(self.__query_contacts(query))
        # query groups
        for group in self.__query_groups(query):
            for contact in group[1]:
                if contact not in matching_contacts:
                    matching_contacts.append(contact)

        matching_contacts = sorted(matching_contacts, key=lambda c: c.display_name)

        for contact in matching_contacts:
            print("-------------------------", file=out)
            print(contact.display_name, file=out)
            for org in contact.organizations:
                if org.name:
                    print("Organization: ", org.name, file=out)
                if org.department:
                    print("Department:   ", org.department, file=out)
                if org.title:
                    print("Title:        ", org.title, file=out)
            if contact.birthday:
                print("Birthday:     ", contact.birthday.strftime("%x"), file=out)
            if contact.phonenumbers:
                print("Phone:", file=out)
                for (number, kind) in contact.phonenumbers:
                    print("\t", number, " (" + kind + ")", file=out)
            if contact.emails:
                print("EMail:", file=out)
                emailaddrs = sorted(contact.emails)
                for (emailaddr, kind) in emailaddrs:
                    print("\t", emailaddr, " (" + kind + ")", file=out)
            if contact.im:
                print("IM:", file=out)
                for (nick, protocol) in contact.im:
                    print("\t", nick, " (", protocol, ")", file=out)
            if contact.addresses:
                print("Address:", file=out)
                for (address, kind) in contact.addresses:
                    lines = address.splitlines()
                    lines[0] = '%s ( %s )' % (lines[0], kind)
                    print("\t" + '\n\t'.join(lines), file=out)
            groups = set(contact.groups).difference(('My Contacts',))
            if groups:
                print("Groups:", file=out)
                groups_str = '\n\t'.join(groups)
                print("\t" + groups_str, file=out)

    def __query_contacts(self, query):
        match = re.compile(query.replace(' ', '.*'), re.I).search  # create a match function
        for contact in self.cache.contacts:
            if self.__config.filter_groupless_contacts and not contact.groups:
                continue  # Skip contacts without groups
            if any(map(match, list(contact.all_names) + [str(number) for (number, kind) in contact.phonenumbers])):
                yield contact
                continue
            matching_addrs = [(email, kind) for (email, kind) in contact.emails if match(email)]
            if matching_addrs:
                contact.emails = matching_addrs  # only show matching
                yield contact
                continue
            for org in contact.organizations:
                for field in ('name', 'title', 'department'):
                    if org[field] and match(org[field]):
                        yield contact
                        continue

    def __query_groups(self, query):
        match = re.compile(query.replace(' ', '.*'), re.I).search  # create a match function
        for group in self.cache.groups:
            # Collect all values to match against
            all_values = (group,)
            if any(map(match, all_values)):
                contacts = list(self.__get_group_contacts(group))
                yield group, contacts

    def __get_group_contacts(self, group):
        for contact in self.cache.contacts:
            if group in contact.groups:
                yield contact

    def add_mail_contact(self, name, mailaddr, phone=None):
        contact = {
            'names': [{'givenName': name}],
            'emailAddresses': [{'value': mailaddr}],
            'phoneNumbers': [{'value': phone}],
        }

        gcont = GoogleContacts(self.__config)
        log.debug('Going to create contact name: %s email: %s phone: %s', name, mailaddr, phone)
        gcont.create_contact(contact)
        log.info('Created contact name: %s email: %s %s', name, mailaddr, phone)

    def add_email_from(self, lines):
        """Add an address from From: field of a mail.

        This assumes a single mail file is supplied through.

        Args:
        ----
          lines: A generator of lines, usually a open file.

        """
        parser = email.parser.HeaderParser()
        headers = parser.parse(lines)
        if 'From' not in headers:
            print("Not a valid mail file!")
            sys.exit(2)

        (name, mailaddr) = email.utils.parseaddr(headers['From'])
        if not name:
            name = mailaddr
        else:
            # This decodes headers like "=?iso-8859-1?q?p=F6stal?="
            values = email.header.decode_header(name)
            if not values:
                # Can't this be possible?
                name = mailaddr
            else:
                # There should be only one element anyway
                (name, encoding) = values[0]

                if encoding is not None:
                    name = name.decode(encoding)

        self.add_mail_contact(name, mailaddr)


class Cache():
    def __init__(self, config):
        self.__config = config
        self.contacts = None  # list of Storage
        self.groups = None  # list of Storage

    def load(self, force_update=False):
        """Load the cached addressbook feed, or fetch it (again) if it is old or missing or invalid or anything.

        Args:
        ----
          force_update: force update of cache

        """
        cache = {}

        # if cache newer than cache_expiry_hours
        if not force_update and (os.path.exists(self.__config.cache_filename) and
                                 ((time.time() - os.path.getmtime(self.__config.cache_filename)) <
                                  (int(self.__config.cache_expiry_hours) * 60 * 60))):
            try:
                log.debug('Loading cache: %s', self.__config.cache_filename)
                cache = pickle.load(open(self.__config.cache_filename, 'rb'))
                if cache.get('goobook_cache') != CACHE_FORMAT_VERSION:
                    log.info('Detected old cache format')
                    cache = None  # Old cache format
            except Exception as err:
                log.info('Failed to read the cache file: %s', err)
                raise
        if cache:
            self.contacts = storageify(cache.get('contacts'))
            self.groups = storageify(cache.get('groups'))
        else:
            self.update()
        if not self.contacts:
            raise Exception('Failed to find any contacts')  # TODO

    def update(self):
        log.info('Retrieving contact data from Google.')
        gcs = GoogleContacts(self.__config)
        groupname_by_id = parse_groups(gcs.fetch_contact_groups())
        self.contacts = list(parse_contacts(gcs.fetch_contacts(), groupname_by_id))
        self.groups = list(groupname_by_id.values())
        self.save()

    def save(self):
        """Pickle the addressbook and a timestamp."""
        if self.contacts:  # never write a empty addressbook
            cache = {'contacts': unstorageify(self.contacts),
                     'groups': unstorageify(self.groups),
                     'goobook_cache': CACHE_FORMAT_VERSION}
            pickle.dump(cache, open(self.__config.cache_filename, 'wb'))


def parse_contact(person, groupname_by_id):
    """Extracts interesting contact info from cache.

    https://developers.google.com/people/api/rest/v1/people
    """
    contact = Storage()
    contact.emails = []
    contact.birthday = None  # datetime.date
    contact.im = []
    contact.addresses = []
    contact.display_name = None
    contact.all_names = []
    contact.groups = []
    contact.phonenumbers = []  # [TypedValue]
    contact.organizations = []  # [Storage()]

    for emaila in person.get('emailAddresses', []):
        contact.emails.append(TypedValue(emaila['value'], emaila.get('type', '')))

    if 'birthdays' in person.keys() and person['birthdays'] and 'date' in person['birthdays'][0]:
        birthday = person['birthdays'][0]['date']
        if len(birthday) == 3:  # we skip incomplete birthdates
            contact.birthday = datetime.date(birthday['year'], birthday['month'], birthday['day'])

    for name in person.get('names', []):
        if 'displayName' in name and contact.display_name is None:
            # use first displayName found
            contact.display_name = name['displayName']
        for field in ("displayName", "displayNameLastFirst", "familyName", "givenName", "middleName",
                      "honorificPrefix", "honorificSuffix", "phoneticFullName", "phoneticFamilyName",
                      "phoneticGivenName", "phoneticMiddleName", "phoneticHonorificPrefix",
                      "phoneticHonorificSuffix"):
            if field in name:
                contact.all_names.append(name[field])

    if contact.display_name is None and contact.emails:
        # if there is no displayName use a email address
        contact.display_name = contact.emails[0].value

    for membership in person.get('memberships', []):
        if "contactGroupMembership" in membership:
            contact.groups.append(groupname_by_id['contactGroups/' +
                                                  membership['contactGroupMembership']['contactGroupId']])

    for phone in person.get('phoneNumbers', []):
        contact.phonenumbers.append(TypedValue(phone['value'], phone.get('type', '')))

    for address in person.get('addresses', []):
        if 'formattedValue' in address:
            contact.addresses.append(TypedValue(address['formattedValue'], address.get('type', '')))

    for item in person.get('imClients', []):
        if 'username' in item:
            contact.im.append(TypedValue(item['username'], item.get('protocol', '')))

    if 'organizations' in person.keys() and person['organizations']:
        for org in person['organizations']:
            if not contact.display_name:
                contact.display_name = org.get('name')
            contact.organizations.append(Storage(name=org.get('name'),
                                                 title=org.get('title'),
                                                 department=org.get('department')))
    log.debug('Parsed contact %s', contact)
    if not contact.display_name:
        log.info('Skipping contact because of no name/e-mail/organization: %s', person)
        return None
    return contact


def parse_contacts(raw_contacts, groupname_by_id):
    for contact in raw_contacts:
        parsed = parse_contact(contact, groupname_by_id)
        if parsed:
            yield parsed


def parse_groups(raw_groups):
    groupname_by_id = {}
    for entry in raw_groups:
        groupname_by_id[entry['resourceName']] = entry['formattedName']
    return groupname_by_id


class GoogleContacts():

    def __init__(self, config):
        http_client = self.__get_client(config.creds)
        self.service = build('people', 'v1', http=http_client)
        # self.__additional_headers = {
        #     'GData-Version': GDATA_VERSION,
        #     'Content-Type': 'application/atom+xml'
        # }

    @staticmethod
    def __get_client(credentials):
        """Login to Google and return a ContactsClient object."""
        if not credentials or credentials.invalid:
            sys.exit('No or invalid credentials, run "goobook authenticate"')  # TODO raise exception instead
        http_auth = credentials.authorize(httplib2.Http())
        return http_auth

    def fetch_contacts(self):
        connections = []
        request = self.service.people().connections().list(
            resourceName='people/me',
            pageSize=2000,  # Number of connections in response
            x__xgafv=None,
            pageToken=None,
            sortOrder=None,
            personFields=('names,nicknames,emailAddresses,memberships,'
                          'phoneNumbers,birthdays,imClients,organizations,addresses'),
            requestSyncToken=None,
            syncToken=None,
            requestMask_includeField=None)

        # Loop until all pages have been processed.
        while request is not None:
            # Get the next page.
            response = request.execute()
            # Accessing the response like a dict object with an 'items' key
            # returns a list of item objects (connections).
            connections.extend(response.get('connections', []))
            # Get the next request object by passing the previous request object to
            # the list_next method.
            request = self.service.people().connections().list_next(request, response)
        return connections

    def fetch_contact_groups(self):
        groups = []
        request = self.service.contactGroups().list(pageSize=500)

        # Loop until all pages have been processed.
        while request is not None:
            # Get the next page.
            response = request.execute()
            # Accessing the response like a dict object with an 'items' key
            # returns a list of item objects (groups).
            groups.extend(response.get('contactGroups', []))
            # Get the next request object by passing the previous request object to
            # the list_next method.
            request = self.service.contactGroups().list_next(request, response)
        return groups

    def create_contact(self, contact):
        self.service.people().createContact(body=contact).execute()



#!/usr/bin/env python
# vim: fileencoding=UTF-8 filetype=python ff=unix expandtab sw=4 sts=4 tw=120
# author: Christer Sjöholm -- goobook AT furuvik DOT net

import os
import setuptools

HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
NEWS = open(os.path.join(HERE, 'CHANGES.rst')).read()


setuptools.setup(
    name='goobook',
    version='3.5.1',
    description='Search your google contacts from the command-line or mutt.',
    long_description=README + '\n\n' + NEWS,
    long_description_content_type="text/x-rst",
    author='Christer Sjöholm',
    author_email='goobook@furuvik.net',
    url='http://gitlab.com/goobook/goobook',
    download_url='http://pypi.python.org/pypi/goobook',
    classifiers=[f.strip() for f in """
        Development Status :: 5 - Production/Stable
        Environment :: Console
        Operating System :: OS Independent
        Programming Language :: Python
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3.7
        Programming Language :: Python :: 3.8
        Intended Audience :: End Users/Desktop
        License :: OSI Approved :: GNU General Public License v3 (GPLv3)
        Topic :: Communications :: Email :: Address Book
    """.splitlines() if f.strip()],
    keywords='abook mutt e-mail gmail google address-book',
    license='GPLv3',
    install_requires=[
        'google-api-python-client>=1.7.12',
        'simplejson>=3.16.0',
        'oauth2client>=1.5.0,<5.0.0dev',
        'xdg>=4.0.1'
    ],
    extras_require={
    },
    include_package_data=True,
    zip_safe=False,
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['goobook = goobook.application:main']}
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# vim: fileencoding=UTF-8 filetype=python ff=unix et ts=4 sw=4 sts=4 tw=120
# author: Christer Sjöholm -- hcs AT furuvik DOT net


import argparse
import datetime
import locale
import logging
import pathlib
import json
import sys

import oauth2client.client
import oauth2client.file
import oauth2client.tools
import pkg_resources

import goobook.config
from goobook.goobook import GooBook, Cache, GoogleContacts, parse_groups, parse_contacts
from goobook.storage import unstorageify

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

SCOPES = 'https://www.google.com/m8/feeds'  # read/write access to Contacts and Contact Groups

AUTHENTICATE_HELP_STRING = '''Google OAuth authentication.

Before running goobook authenticate you need a client_id and a client_secret, get it like this:

Go to https://developers.google.com/people/quickstart/python
and click "Enable the People API"
select a name (ex. GooBook)
select desktop application
save the client_id and client_secret to be used below::

    $ goobook authenticate -- CLIENT_ID CLIENT_SECRET

and follow the instructions.

if it doesn't open a webbrowser use

goobook authenticate --noauth_local_webserver -- CLIENT_ID CLIENT_SECRET

If you get the page "This app isn't verified" select Advanced and the "Go to GooBook (unsafe)" at the bottom.
'''


def main():
    locale.setlocale(locale.LC_TIME, '')  # Use system configured locale

    parser = argparse.ArgumentParser(description='Search you Google contacts from mutt or the command-line.')
    parser.add_argument('-c', '--config', help='Specify alternative configuration file.', metavar="FILE")
    parser.add_argument('-v', '--verbose', dest="log_level", action='store_const',
                        const=logging.INFO, help='Be verbose about what is going on (stderr).')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%%(prog)s %s' % pkg_resources.get_distribution("goobook").version,
                        help="Print version and exit")
    parser.add_argument('-d', '--debug', dest="log_level", action='store_const',
                        const=logging.DEBUG, help='Output debug info (stderr).')
    parser.set_defaults(log_level=logging.ERROR)

    subparsers = parser.add_subparsers()

    parser_add = subparsers.add_parser('add',
                                       description='Create new contact, if name and email is not given the'
                                       ' sender of a mail read from stdin will be used.')
    parser_add.add_argument('name', nargs='?', metavar='NAME',
                            help='Name to use.')
    parser_add.add_argument('email', nargs='?', metavar='EMAIL',
                            help='E-mail to use.')
    parser_add.add_argument('phone', nargs='?', metavar='PHONE',
                            help='Phone number to use.')
    parser_add.set_defaults(func=do_add)

    parser_config_template = subparsers.add_parser('config-template',
                                                   description='Prints a template for .goobookrc to stdout')
    parser_config_template.set_defaults(func=do_config_template)

    parser_dump_contacts = subparsers.add_parser('dump_contacts',
                                                 description='Dump contacts as JSON.')
    parser_dump_contacts.add_argument('-p', '--parse', action='store_true', help='Dump parsed contact instead of raw.')
    parser_dump_contacts.set_defaults(func=do_dump_contacts)

    parser_dump_groups = subparsers.add_parser('dump_groups',
                                               description='Dump groups as JSON.')
    parser_dump_groups.add_argument('-p', '--parse', action='store_true', help='Dump parsed contact instead of raw.')
    parser_dump_groups.set_defaults(func=do_dump_groups)

    parser_query = subparsers.add_parser('query',
                                         description='Search contacts using query (regex).')
    parser_query.add_argument('-s', '--simple', action='store_true',
                                    help='Simple output format instead of mutt compatible')
    parser_query.add_argument('query', help='regex to search for.', metavar='QUERY')
    parser_query.set_defaults(func=do_query)

    parser_query_details = subparsers.add_parser(
        'dquery',
        description='Search contacts using query (regex) and print out all info.')
    parser_query_details.add_argument('query', help='regex to search for.')
    parser_query_details.set_defaults(func=do_query_details)

    parser_reload = subparsers.add_parser('reload',
                                          description='Force reload of the cache.')
    parser_reload.set_defaults(func=do_reload)

    parser_auth = subparsers.add_parser('authenticate',
                                        description=AUTHENTICATE_HELP_STRING,
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        parents=[oauth2client.tools.argparser])
    parser_auth.add_argument('client_id', metavar='CLIENT_ID',
                             help='Client ID')
    parser_auth.add_argument('client_secret', metavar='CLIENT_SECRET',
                             help='Client secret')
    parser_auth.set_defaults(func=do_authenticate)

    parser_unauth = subparsers.add_parser('unauthenticate',
                                          description="Removed authentication data (logout).")
    parser_unauth.set_defaults(func=do_unauthenticate)

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level)

    if 'func' not in args:
        parser.error('To few arguments.')

    try:
        if args.func is do_config_template:
            config = None
        else:
            config = goobook.config.read_config(args.config)
        args.func(config, args)
    except goobook.config.ConfigError as err:
        sys.exit('Configuration error: ' + str(err))

##############################################################################
# sub commands


def do_add(config, args):
    goobk = GooBook(config)
    if args.name and args.email:
        goobk.add_mail_contact(args.name, args.email, args.phone)
    else:
        goobk.add_email_from(sys.stdin)
    goobk.cache.load(force_update=True)


def do_config_template(_config, _args):
    print(goobook.config.TEMPLATE)


def do_dump_contacts(config, args):
    goco = GoogleContacts(config)
    contacts = goco.fetch_contacts()
    if args.parse:
        groupname_by_id = parse_groups(goco.fetch_contact_groups())
        contacts = unstorageify(list(parse_contacts(goco.fetch_contacts(), groupname_by_id)))

    class DateEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime.date):
                return str(obj)
            # Let the base class default method raise the TypeError
            return json.JSONEncoder.default(self, obj)
    print(json.dumps(contacts, sort_keys=True, indent=4, ensure_ascii=False, cls=DateEncoder))


def do_dump_groups(config, args):
    goco = GoogleContacts(config)
    groups = goco.fetch_contact_groups()
    if args.parse:
        groupname_by_id = parse_groups(goco.fetch_contact_groups())
        groups = unstorageify(list(groupname_by_id.values()))

    print(json.dumps(groups, sort_keys=True, indent=4, ensure_ascii=False))


def do_query(config, args):
    goobk = GooBook(config)
    goobk.query(args.query, simple=args.simple)


def do_query_details(config, args):
    goobk = GooBook(config)
    goobk.query_details(args.query)


def do_reload(config, _args):
    cache = Cache(config)
    cache.load(force_update=True)


def do_authenticate(config, args):
    store = config.store
    creds = config.creds

    if not creds or creds.invalid:
        flow = oauth2client.client.OAuth2WebServerFlow(args.client_id, args.client_secret, SCOPES)
        creds = oauth2client.tools.run_flow(flow, store, args)
    else:
        print('You are already authenticated.')


def do_unauthenticate(config, _args):
    oauth_db = pathlib.Path(config.oauth_db_filename)
    if oauth_db.exists():
        oauth_db.unlink()
        print("deleted", oauth_db)


if __name__ == '__main__':
    main()
